/* We don't want gtkdoc scanning this file, it'll get
 * confused by seeing function definitions, so mark it as: */
/*<private_header>*/

/**
 * emp_cli_channel_handler_callback_for_handle_channel:
 * @proxy: the proxy on which the call was made
 * @error: %NULL on success, or an error on failure
 * @user_data: user-supplied data
 * @weak_object: user-supplied object
 *
 * Signature of the callback called when a HandleChannel method call
 * succeeds or fails.
 */
static void
_emp_cli_channel_handler_collect_callback_handle_channel (DBusGProxy *proxy,
    DBusGProxyCall *call,
    gpointer user_data)
{
  GError *error = NULL;

  dbus_g_proxy_end_call (proxy, call, &error,
      G_TYPE_INVALID);
  tp_proxy_pending_call_v0_take_results (user_data, error,NULL);
}
static void
_emp_cli_channel_handler_invoke_callback_handle_channel (TpProxy *self,
    GError *error,
    GValueArray *args,
    GCallback generic_callback,
    gpointer user_data,
    GObject *weak_object)
{
  emp_cli_channel_handler_callback_for_handle_channel callback = (emp_cli_channel_handler_callback_for_handle_channel) generic_callback;

  if (error != NULL)
    {
      callback ((TpProxy *) self,
          error, user_data, weak_object);
      g_error_free (error);
      return;
    }
  callback ((TpProxy *) self,
      error, user_data, weak_object);

  if (args != NULL)
    g_value_array_free (args);
}

/**
 * emp_cli_channel_handler_call_handle_channel:
 * @proxy: the #TpProxy
 * @timeout_ms: the timeout in milliseconds, or -1 to use the
 *   default
 * @in_Bus_Name: Used to pass an 'in' argument: The bus name of the connection and channel
 * @in_Connection: Used to pass an 'in' argument: The object-path of the connection that owns the channel
 * @in_Channel_Type: Used to pass an 'in' argument: The channel type
 * @in_Channel: Used to pass an 'in' argument: The object-path of the channel
 * @in_Handle_Type: Used to pass an 'in' argument: The type of the handle that the channel communicates           with, or 0 if there is no associated handle
 * @in_Handle: Used to pass an 'in' argument: The handle that the channel communicates with,           or 0 if there is no associated handle
 * @callback: called when the method call succeeds or fails;
 *   may be %NULL to make a "fire and forget" call with no 
 *   reply tracking
 * @user_data: user-supplied data passed to the callback;
 *   must be %NULL if @callback is %NULL
 * @destroy: called with the user_data as argument, after the
 *   call has succeeded, failed or been cancelled;
 *   must be %NULL if @callback is %NULL
 * @weak_object: If not %NULL, a #GObject which will be 
 *   weakly referenced; if it is destroyed, this call 
 *   will automatically be cancelled. Must be %NULL if 
 *   @callback is %NULL
 *
 * Start a HandleChannel method call.
 *
 * Called when a channel handler should handle a new channel.
 *
 * Returns: a #TpProxyPendingCall representing the call in
 *  progress. It is borrowed from the object, and will become
 *  invalid when the callback is called, the call is
 *  cancelled or the #TpProxy becomes invalid.
 */
TpProxyPendingCall *
emp_cli_channel_handler_call_handle_channel (TpProxy *proxy,
    gint timeout_ms,
    const gchar *in_Bus_Name,
    const gchar *in_Connection,
    const gchar *in_Channel_Type,
    const gchar *in_Channel,
    guint in_Handle_Type,
    guint in_Handle,
    emp_cli_channel_handler_callback_for_handle_channel callback,
    gpointer user_data,
    GDestroyNotify destroy,
    GObject *weak_object)
{
  GError *error = NULL;
  GQuark interface = EMP_IFACE_QUARK_CHANNEL_HANDLER;
  DBusGProxy *iface;

  g_return_val_if_fail (TP_IS_PROXY (proxy), NULL);
  g_return_val_if_fail (callback != NULL || user_data == NULL, NULL);
  g_return_val_if_fail (callback != NULL || destroy == NULL, NULL);
  g_return_val_if_fail (callback != NULL || weak_object == NULL, NULL);

  iface = tp_proxy_borrow_interface_by_id (
      (TpProxy *) proxy,
      interface, &error);

  if (iface == NULL)
    {
      if (callback != NULL)
        callback (proxy,
            error, user_data, weak_object);

      if (destroy != NULL)
        destroy (user_data);

      g_error_free (error);
      return NULL;
    }

  if (callback == NULL)
    {
      dbus_g_proxy_call_no_reply (iface, "HandleChannel",
          G_TYPE_STRING, in_Bus_Name,
          DBUS_TYPE_G_OBJECT_PATH, in_Connection,
          G_TYPE_STRING, in_Channel_Type,
          DBUS_TYPE_G_OBJECT_PATH, in_Channel,
          G_TYPE_UINT, in_Handle_Type,
          G_TYPE_UINT, in_Handle,
          G_TYPE_INVALID);
      return NULL;
    }
  else
    {
      TpProxyPendingCall *data;

      data = tp_proxy_pending_call_v0_new ((TpProxy *) proxy,
          interface, "HandleChannel", iface,
          _emp_cli_channel_handler_invoke_callback_handle_channel,
          G_CALLBACK (callback), user_data, destroy,
          weak_object, FALSE);
      tp_proxy_pending_call_v0_take_pending_call (data,
          dbus_g_proxy_begin_call_with_timeout (iface,
              "HandleChannel",
              _emp_cli_channel_handler_collect_callback_handle_channel,
              data,
              tp_proxy_pending_call_v0_completed,
              timeout_ms,
              G_TYPE_STRING, in_Bus_Name,
              DBUS_TYPE_G_OBJECT_PATH, in_Connection,
              G_TYPE_STRING, in_Channel_Type,
              DBUS_TYPE_G_OBJECT_PATH, in_Channel,
              G_TYPE_UINT, in_Handle_Type,
              G_TYPE_UINT, in_Handle,
              G_TYPE_INVALID));

      return data;
    }
}

typedef struct {
    GMainLoop *loop;
    GError **error;
    unsigned success:1;
    unsigned completed:1;
} _emp_cli_channel_handler_run_state_handle_channel;
static void
_emp_cli_channel_handler_finish_running_handle_channel (TpProxy *self G_GNUC_UNUSED,
    GError *error,
    GValueArray *args,
    GCallback unused G_GNUC_UNUSED,
    gpointer user_data G_GNUC_UNUSED,
    GObject *unused2 G_GNUC_UNUSED)
{
  _emp_cli_channel_handler_run_state_handle_channel *state = user_data;

  state->success = (error == NULL);
  state->completed = TRUE;
  g_main_loop_quit (state->loop);

  if (error != NULL)
    {
      if (state->error != NULL)
        *state->error = error;
      else
        g_error_free (error);

      return;
    }

  if (args != NULL)
    g_value_array_free (args);
}

/**
 * emp_cli_channel_handler_run_handle_channel:
 * @proxy: A #TpProxy or subclass
 * @timeout_ms: Timeout in milliseconds, or -1 for default
 * @in_Bus_Name: Used to pass an 'in' argument: The bus name of the connection and channel
 * @in_Connection: Used to pass an 'in' argument: The object-path of the connection that owns the channel
 * @in_Channel_Type: Used to pass an 'in' argument: The channel type
 * @in_Channel: Used to pass an 'in' argument: The object-path of the channel
 * @in_Handle_Type: Used to pass an 'in' argument: The type of the handle that the channel communicates           with, or 0 if there is no associated handle
 * @in_Handle: Used to pass an 'in' argument: The handle that the channel communicates with,           or 0 if there is no associated handle
 * @error: If not %NULL, used to return errors if %FALSE 
 *  is returned
 * @loop: If not %NULL, set before re-entering 
 *  the main loop, to point to a #GMainLoop 
 *  which can be used to cancel this call with 
 *  g_main_loop_quit(), causing a return of 
 *  %FALSE with @error set to %TP_DBUS_ERROR_CANCELLED
 *
 * Call the method HandleChannel and run the main loop
 * until it returns. Before calling this method, you must
 * add a reference to any borrowed objects you need to keep,
 * and generally ensure that everything is in a consistent
 * state.
 *
 * Called when a channel handler should handle a new channel.
 *
 * Returns: TRUE on success, FALSE and sets @error on error
 */
gboolean
emp_cli_channel_handler_run_handle_channel (TpProxy *proxy,
    gint timeout_ms,
    const gchar *in_Bus_Name,
    const gchar *in_Connection,
    const gchar *in_Channel_Type,
    const gchar *in_Channel,
    guint in_Handle_Type,
    guint in_Handle,
    GError **error,
    GMainLoop **loop)
{
  DBusGProxy *iface;
  GQuark interface = EMP_IFACE_QUARK_CHANNEL_HANDLER;
  TpProxyPendingCall *pc;
  _emp_cli_channel_handler_run_state_handle_channel state = {
      NULL /* loop */, error,
      FALSE /* completed */, FALSE /* success */ };

  g_return_val_if_fail (TP_IS_PROXY (proxy), FALSE);

  iface = tp_proxy_borrow_interface_by_id
       ((TpProxy *) proxy, interface, error);

  if (iface == NULL)
    return FALSE;

  state.loop = g_main_loop_new (NULL, FALSE);

  pc = tp_proxy_pending_call_v0_new ((TpProxy *) proxy,
      interface, "HandleChannel", iface,
      _emp_cli_channel_handler_finish_running_handle_channel,
      NULL, &state, NULL, NULL, TRUE);

  if (loop != NULL)
    *loop = state.loop;

  tp_proxy_pending_call_v0_take_pending_call (pc,
      dbus_g_proxy_begin_call_with_timeout (iface,
          "HandleChannel",
          _emp_cli_channel_handler_collect_callback_handle_channel,
          pc,
          tp_proxy_pending_call_v0_completed,
          timeout_ms,
              G_TYPE_STRING, in_Bus_Name,
              DBUS_TYPE_G_OBJECT_PATH, in_Connection,
              G_TYPE_STRING, in_Channel_Type,
              DBUS_TYPE_G_OBJECT_PATH, in_Channel,
              G_TYPE_UINT, in_Handle_Type,
              G_TYPE_UINT, in_Handle,
          G_TYPE_INVALID));

  if (!state.completed)
    g_main_loop_run (state.loop);

  if (!state.completed)
    tp_proxy_pending_call_cancel (pc);

  if (loop != NULL)
    *loop = NULL;

  g_main_loop_unref (state.loop);

  return state.success;
}


static inline void
emp_cli_add_signals_for_debug (DBusGProxy *proxy)
{
  if (!tp_proxy_dbus_g_proxy_claim_for_signal_adding (proxy))
    return;
  dbus_g_proxy_add_signal (proxy, "NewDebugMessage",
      G_TYPE_DOUBLE,
      G_TYPE_STRING,
      G_TYPE_UINT,
      G_TYPE_STRING,
      G_TYPE_INVALID);
}


/**
 * emp_cli_debug_signal_callback_new_debug_message:
 * @proxy: The proxy on which emp_cli_debug_connect_to_new_debug_message ()
 *  was called
 * @arg_time: Timestamp of the debug message.
 * @arg_domain: Domain of the debug message, as described in the Debug_Message struct.
 * @arg_level: Level of the debug message.
 * @arg_message: The text of the debug message.
 * @user_data: User-supplied data
 * @weak_object: User-supplied weakly referenced object
 *
 * Represents the signature of a callback for the signal NewDebugMessage.
 */
static void
_emp_cli_debug_collect_args_of_new_debug_message (DBusGProxy *proxy G_GNUC_UNUSED,
    gdouble arg_time,
    const gchar *arg_domain,
    guint arg_level,
    const gchar *arg_message,
    TpProxySignalConnection *sc)
{
  GValueArray *args = g_value_array_new (4);
  GValue blank = { 0 };
  guint i;

  g_value_init (&blank, G_TYPE_INT);

  for (i = 0; i < 4; i++)
    g_value_array_append (args, &blank);

  g_value_unset (args->values + 0);
  g_value_init (args->values + 0, G_TYPE_DOUBLE);
  g_value_set_double (args->values + 0, arg_time);

  g_value_unset (args->values + 1);
  g_value_init (args->values + 1, G_TYPE_STRING);
  g_value_set_string (args->values + 1, arg_domain);

  g_value_unset (args->values + 2);
  g_value_init (args->values + 2, G_TYPE_UINT);
  g_value_set_uint (args->values + 2, arg_level);

  g_value_unset (args->values + 3);
  g_value_init (args->values + 3, G_TYPE_STRING);
  g_value_set_string (args->values + 3, arg_message);

  tp_proxy_signal_connection_v0_take_results (sc, args);
}
static void
_emp_cli_debug_invoke_callback_for_new_debug_message (TpProxy *tpproxy,
    GError *error G_GNUC_UNUSED,
    GValueArray *args,
    GCallback generic_callback,
    gpointer user_data,
    GObject *weak_object)
{
  emp_cli_debug_signal_callback_new_debug_message callback =
      (emp_cli_debug_signal_callback_new_debug_message) generic_callback;

  if (callback != NULL)
    callback (g_object_ref (tpproxy),
      g_value_get_double (args->values + 0),
      g_value_get_string (args->values + 1),
      g_value_get_uint (args->values + 2),
      g_value_get_string (args->values + 3),
      user_data,
      weak_object);

  g_value_array_free (args);
  g_object_unref (tpproxy);
}
/**
 * emp_cli_debug_connect_to_new_debug_message:
 * @proxy: A #TpProxy or subclass
 * @callback: Callback to be called when the signal is
 *   received
 * @user_data: User-supplied data for the callback
 * @destroy: Destructor for the user-supplied data, which
 *   will be called when this signal is disconnected, or
 *   before this function returns %NULL
 * @weak_object: A #GObject which will be weakly referenced; 
 *   if it is destroyed, this callback will automatically be
 *   disconnected
 * @error: If not %NULL, used to raise an error if %NULL is
 *   returned
 *
 * Connect a handler to the signal NewDebugMessage.
 *
 * Emitted when a debug messages is generated if the         &lt;tp:member-ref&gt;Enabled&lt;/tp:member-ref&gt; property is set to TRUE.
 *
 * Returns: a #TpProxySignalConnection containing all of the
 * above, which can be used to disconnect the signal; or
 * %NULL if the proxy does not have the desired interface
 * or has become invalid.
 */
TpProxySignalConnection *
emp_cli_debug_connect_to_new_debug_message (TpProxy *proxy,
    emp_cli_debug_signal_callback_new_debug_message callback,
    gpointer user_data,
    GDestroyNotify destroy,
    GObject *weak_object,
    GError **error)
{
  GType expected_types[5] = {
      G_TYPE_DOUBLE,
      G_TYPE_STRING,
      G_TYPE_UINT,
      G_TYPE_STRING,
      G_TYPE_INVALID };

  g_return_val_if_fail (TP_IS_PROXY (proxy), NULL);
  g_return_val_if_fail (callback != NULL, NULL);

  return tp_proxy_signal_connection_v0_new ((TpProxy *) proxy,
      EMP_IFACE_QUARK_DEBUG, "NewDebugMessage",
      expected_types,
      G_CALLBACK (_emp_cli_debug_collect_args_of_new_debug_message),
      _emp_cli_debug_invoke_callback_for_new_debug_message,
      G_CALLBACK (callback), user_data, destroy,
      weak_object, error);
}

/**
 * emp_cli_debug_callback_for_get_messages:
 * @proxy: the proxy on which the call was made
 * @out_Messages: Used to return an 'out' argument if @error is %NULL: A list of debug messages.
 * @error: %NULL on success, or an error on failure
 * @user_data: user-supplied data
 * @weak_object: user-supplied object
 *
 * Signature of the callback called when a GetMessages method call
 * succeeds or fails.
 */
static void
_emp_cli_debug_collect_callback_get_messages (DBusGProxy *proxy,
    DBusGProxyCall *call,
    gpointer user_data)
{
  GError *error = NULL;
  GValueArray *args;
  GValue blank = { 0 };
  guint i;
  GPtrArray *out_Messages;

  dbus_g_proxy_end_call (proxy, call, &error,
      (dbus_g_type_get_collection ("GPtrArray", (dbus_g_type_get_struct ("GValueArray", G_TYPE_DOUBLE, G_TYPE_STRING, G_TYPE_UINT, G_TYPE_STRING, G_TYPE_INVALID)))), &out_Messages,
      G_TYPE_INVALID);

  if (error != NULL)
    {
      tp_proxy_pending_call_v0_take_results (user_data, error,
          NULL);
      return;
    }

  args = g_value_array_new (1);
  g_value_init (&blank, G_TYPE_INT);

  for (i = 0; i < 1; i++)
    g_value_array_append (args, &blank);

  g_value_unset (args->values + 0);
  g_value_init (args->values + 0, (dbus_g_type_get_collection ("GPtrArray", (dbus_g_type_get_struct ("GValueArray", G_TYPE_DOUBLE, G_TYPE_STRING, G_TYPE_UINT, G_TYPE_STRING, G_TYPE_INVALID)))));
  g_value_take_boxed (args->values + 0, out_Messages);
  tp_proxy_pending_call_v0_take_results (user_data, NULL, args);
}
static void
_emp_cli_debug_invoke_callback_get_messages (TpProxy *self,
    GError *error,
    GValueArray *args,
    GCallback generic_callback,
    gpointer user_data,
    GObject *weak_object)
{
  emp_cli_debug_callback_for_get_messages callback = (emp_cli_debug_callback_for_get_messages) generic_callback;

  if (error != NULL)
    {
      callback ((TpProxy *) self,
          NULL,
          error, user_data, weak_object);
      g_error_free (error);
      return;
    }
  callback ((TpProxy *) self,
      g_value_get_boxed (args->values + 0),
      error, user_data, weak_object);

  g_value_array_free (args);
}

/**
 * emp_cli_debug_call_get_messages:
 * @proxy: the #TpProxy
 * @timeout_ms: the timeout in milliseconds, or -1 to use the
 *   default
 * @callback: called when the method call succeeds or fails;
 *   may be %NULL to make a "fire and forget" call with no 
 *   reply tracking
 * @user_data: user-supplied data passed to the callback;
 *   must be %NULL if @callback is %NULL
 * @destroy: called with the user_data as argument, after the
 *   call has succeeded, failed or been cancelled;
 *   must be %NULL if @callback is %NULL
 * @weak_object: If not %NULL, a #GObject which will be 
 *   weakly referenced; if it is destroyed, this call 
 *   will automatically be cancelled. Must be %NULL if 
 *   @callback is %NULL
 *
 * Start a GetMessages method call.
 *
 * Retrieve buffered debug messages. An implementation could have a         limit on how many message it keeps and so the array returned from         this method should not be assumed to be all of the messages in         the lifetime of the service.
 *
 * Returns: a #TpProxyPendingCall representing the call in
 *  progress. It is borrowed from the object, and will become
 *  invalid when the callback is called, the call is
 *  cancelled or the #TpProxy becomes invalid.
 */
TpProxyPendingCall *
emp_cli_debug_call_get_messages (TpProxy *proxy,
    gint timeout_ms,
    emp_cli_debug_callback_for_get_messages callback,
    gpointer user_data,
    GDestroyNotify destroy,
    GObject *weak_object)
{
  GError *error = NULL;
  GQuark interface = EMP_IFACE_QUARK_DEBUG;
  DBusGProxy *iface;

  g_return_val_if_fail (TP_IS_PROXY (proxy), NULL);
  g_return_val_if_fail (callback != NULL || user_data == NULL, NULL);
  g_return_val_if_fail (callback != NULL || destroy == NULL, NULL);
  g_return_val_if_fail (callback != NULL || weak_object == NULL, NULL);

  iface = tp_proxy_borrow_interface_by_id (
      (TpProxy *) proxy,
      interface, &error);

  if (iface == NULL)
    {
      if (callback != NULL)
        callback (proxy,
            NULL,
            error, user_data, weak_object);

      if (destroy != NULL)
        destroy (user_data);

      g_error_free (error);
      return NULL;
    }

  if (callback == NULL)
    {
      dbus_g_proxy_call_no_reply (iface, "GetMessages",
          G_TYPE_INVALID);
      return NULL;
    }
  else
    {
      TpProxyPendingCall *data;

      data = tp_proxy_pending_call_v0_new ((TpProxy *) proxy,
          interface, "GetMessages", iface,
          _emp_cli_debug_invoke_callback_get_messages,
          G_CALLBACK (callback), user_data, destroy,
          weak_object, FALSE);
      tp_proxy_pending_call_v0_take_pending_call (data,
          dbus_g_proxy_begin_call_with_timeout (iface,
              "GetMessages",
              _emp_cli_debug_collect_callback_get_messages,
              data,
              tp_proxy_pending_call_v0_completed,
              timeout_ms,
              G_TYPE_INVALID));

      return data;
    }
}

typedef struct {
    GMainLoop *loop;
    GError **error;
    GPtrArray **out_Messages;
    unsigned success:1;
    unsigned completed:1;
} _emp_cli_debug_run_state_get_messages;
static void
_emp_cli_debug_finish_running_get_messages (TpProxy *self G_GNUC_UNUSED,
    GError *error,
    GValueArray *args,
    GCallback unused G_GNUC_UNUSED,
    gpointer user_data G_GNUC_UNUSED,
    GObject *unused2 G_GNUC_UNUSED)
{
  _emp_cli_debug_run_state_get_messages *state = user_data;

  state->success = (error == NULL);
  state->completed = TRUE;
  g_main_loop_quit (state->loop);

  if (error != NULL)
    {
      if (state->error != NULL)
        *state->error = error;
      else
        g_error_free (error);

      return;
    }

  if (state->out_Messages != NULL)
    *state->out_Messages = g_value_dup_boxed (args->values + 0);

  g_value_array_free (args);
}

/**
 * emp_cli_debug_run_get_messages:
 * @proxy: A #TpProxy or subclass
 * @timeout_ms: Timeout in milliseconds, or -1 for default
 * @out_Messages: Used to return an 'out' argument if %TRUE is returned: A list of debug messages.
 * @error: If not %NULL, used to return errors if %FALSE 
 *  is returned
 * @loop: If not %NULL, set before re-entering 
 *  the main loop, to point to a #GMainLoop 
 *  which can be used to cancel this call with 
 *  g_main_loop_quit(), causing a return of 
 *  %FALSE with @error set to %TP_DBUS_ERROR_CANCELLED
 *
 * Call the method GetMessages and run the main loop
 * until it returns. Before calling this method, you must
 * add a reference to any borrowed objects you need to keep,
 * and generally ensure that everything is in a consistent
 * state.
 *
 * Retrieve buffered debug messages. An implementation could have a         limit on how many message it keeps and so the array returned from         this method should not be assumed to be all of the messages in         the lifetime of the service.
 *
 * Returns: TRUE on success, FALSE and sets @error on error
 */
gboolean
emp_cli_debug_run_get_messages (TpProxy *proxy,
    gint timeout_ms,
    GPtrArray **out_Messages,
    GError **error,
    GMainLoop **loop)
{
  DBusGProxy *iface;
  GQuark interface = EMP_IFACE_QUARK_DEBUG;
  TpProxyPendingCall *pc;
  _emp_cli_debug_run_state_get_messages state = {
      NULL /* loop */, error,
    out_Messages,
      FALSE /* completed */, FALSE /* success */ };

  g_return_val_if_fail (TP_IS_PROXY (proxy), FALSE);

  iface = tp_proxy_borrow_interface_by_id
       ((TpProxy *) proxy, interface, error);

  if (iface == NULL)
    return FALSE;

  state.loop = g_main_loop_new (NULL, FALSE);

  pc = tp_proxy_pending_call_v0_new ((TpProxy *) proxy,
      interface, "GetMessages", iface,
      _emp_cli_debug_finish_running_get_messages,
      NULL, &state, NULL, NULL, TRUE);

  if (loop != NULL)
    *loop = state.loop;

  tp_proxy_pending_call_v0_take_pending_call (pc,
      dbus_g_proxy_begin_call_with_timeout (iface,
          "GetMessages",
          _emp_cli_debug_collect_callback_get_messages,
          pc,
          tp_proxy_pending_call_v0_completed,
          timeout_ms,
          G_TYPE_INVALID));

  if (!state.completed)
    g_main_loop_run (state.loop);

  if (!state.completed)
    tp_proxy_pending_call_cancel (pc);

  if (loop != NULL)
    *loop = NULL;

  g_main_loop_unref (state.loop);

  return state.success;
}


/**
 * emp_cli_tube_handler_callback_for_handle_tube:
 * @proxy: the proxy on which the call was made
 * @error: %NULL on success, or an error on failure
 * @user_data: user-supplied data
 * @weak_object: user-supplied object
 *
 * Signature of the callback called when a HandleTube method call
 * succeeds or fails.
 */
static void
_emp_cli_tube_handler_collect_callback_handle_tube (DBusGProxy *proxy,
    DBusGProxyCall *call,
    gpointer user_data)
{
  GError *error = NULL;

  dbus_g_proxy_end_call (proxy, call, &error,
      G_TYPE_INVALID);
  tp_proxy_pending_call_v0_take_results (user_data, error,NULL);
}
static void
_emp_cli_tube_handler_invoke_callback_handle_tube (TpProxy *self,
    GError *error,
    GValueArray *args,
    GCallback generic_callback,
    gpointer user_data,
    GObject *weak_object)
{
  emp_cli_tube_handler_callback_for_handle_tube callback = (emp_cli_tube_handler_callback_for_handle_tube) generic_callback;

  if (error != NULL)
    {
      callback ((TpProxy *) self,
          error, user_data, weak_object);
      g_error_free (error);
      return;
    }
  callback ((TpProxy *) self,
      error, user_data, weak_object);

  if (args != NULL)
    g_value_array_free (args);
}

/**
 * emp_cli_tube_handler_call_handle_tube:
 * @proxy: the #TpProxy
 * @timeout_ms: the timeout in milliseconds, or -1 to use the
 *   default
 * @in_Bus_Name: Used to pass an 'in' argument: The bus name of the connection and channel
 * @in_Connection: Used to pass an 'in' argument: The object-path of the connection that owns the channel
 * @in_Channel: Used to pass an 'in' argument: The object-path of the tubes channel
 * @in_Handle_Type: Used to pass an 'in' argument: The type of the handle that the channel communicates           with, or 0 if there is no associated handle
 * @in_Handle: Used to pass an 'in' argument: The handle that the channel communicates with,           or 0 if there is no associated handle
 * @callback: called when the method call succeeds or fails;
 *   may be %NULL to make a "fire and forget" call with no 
 *   reply tracking
 * @user_data: user-supplied data passed to the callback;
 *   must be %NULL if @callback is %NULL
 * @destroy: called with the user_data as argument, after the
 *   call has succeeded, failed or been cancelled;
 *   must be %NULL if @callback is %NULL
 * @weak_object: If not %NULL, a #GObject which will be 
 *   weakly referenced; if it is destroyed, this call 
 *   will automatically be cancelled. Must be %NULL if 
 *   @callback is %NULL
 *
 * Start a HandleTube method call.
 *
 * Called when a tube handler should handle a new tube within a tubes channel.
 *
 * Returns: a #TpProxyPendingCall representing the call in
 *  progress. It is borrowed from the object, and will become
 *  invalid when the callback is called, the call is
 *  cancelled or the #TpProxy becomes invalid.
 */
TpProxyPendingCall *
emp_cli_tube_handler_call_handle_tube (TpProxy *proxy,
    gint timeout_ms,
    const gchar *in_Bus_Name,
    const gchar *in_Connection,
    const gchar *in_Channel,
    guint in_Handle_Type,
    guint in_Handle,
    emp_cli_tube_handler_callback_for_handle_tube callback,
    gpointer user_data,
    GDestroyNotify destroy,
    GObject *weak_object)
{
  GError *error = NULL;
  GQuark interface = EMP_IFACE_QUARK_TUBE_HANDLER;
  DBusGProxy *iface;

  g_return_val_if_fail (TP_IS_PROXY (proxy), NULL);
  g_return_val_if_fail (callback != NULL || user_data == NULL, NULL);
  g_return_val_if_fail (callback != NULL || destroy == NULL, NULL);
  g_return_val_if_fail (callback != NULL || weak_object == NULL, NULL);

  iface = tp_proxy_borrow_interface_by_id (
      (TpProxy *) proxy,
      interface, &error);

  if (iface == NULL)
    {
      if (callback != NULL)
        callback (proxy,
            error, user_data, weak_object);

      if (destroy != NULL)
        destroy (user_data);

      g_error_free (error);
      return NULL;
    }

  if (callback == NULL)
    {
      dbus_g_proxy_call_no_reply (iface, "HandleTube",
          G_TYPE_STRING, in_Bus_Name,
          DBUS_TYPE_G_OBJECT_PATH, in_Connection,
          DBUS_TYPE_G_OBJECT_PATH, in_Channel,
          G_TYPE_UINT, in_Handle_Type,
          G_TYPE_UINT, in_Handle,
          G_TYPE_INVALID);
      return NULL;
    }
  else
    {
      TpProxyPendingCall *data;

      data = tp_proxy_pending_call_v0_new ((TpProxy *) proxy,
          interface, "HandleTube", iface,
          _emp_cli_tube_handler_invoke_callback_handle_tube,
          G_CALLBACK (callback), user_data, destroy,
          weak_object, FALSE);
      tp_proxy_pending_call_v0_take_pending_call (data,
          dbus_g_proxy_begin_call_with_timeout (iface,
              "HandleTube",
              _emp_cli_tube_handler_collect_callback_handle_tube,
              data,
              tp_proxy_pending_call_v0_completed,
              timeout_ms,
              G_TYPE_STRING, in_Bus_Name,
              DBUS_TYPE_G_OBJECT_PATH, in_Connection,
              DBUS_TYPE_G_OBJECT_PATH, in_Channel,
              G_TYPE_UINT, in_Handle_Type,
              G_TYPE_UINT, in_Handle,
              G_TYPE_INVALID));

      return data;
    }
}

typedef struct {
    GMainLoop *loop;
    GError **error;
    unsigned success:1;
    unsigned completed:1;
} _emp_cli_tube_handler_run_state_handle_tube;
static void
_emp_cli_tube_handler_finish_running_handle_tube (TpProxy *self G_GNUC_UNUSED,
    GError *error,
    GValueArray *args,
    GCallback unused G_GNUC_UNUSED,
    gpointer user_data G_GNUC_UNUSED,
    GObject *unused2 G_GNUC_UNUSED)
{
  _emp_cli_tube_handler_run_state_handle_tube *state = user_data;

  state->success = (error == NULL);
  state->completed = TRUE;
  g_main_loop_quit (state->loop);

  if (error != NULL)
    {
      if (state->error != NULL)
        *state->error = error;
      else
        g_error_free (error);

      return;
    }

  if (args != NULL)
    g_value_array_free (args);
}

/**
 * emp_cli_tube_handler_run_handle_tube:
 * @proxy: A #TpProxy or subclass
 * @timeout_ms: Timeout in milliseconds, or -1 for default
 * @in_Bus_Name: Used to pass an 'in' argument: The bus name of the connection and channel
 * @in_Connection: Used to pass an 'in' argument: The object-path of the connection that owns the channel
 * @in_Channel: Used to pass an 'in' argument: The object-path of the tubes channel
 * @in_Handle_Type: Used to pass an 'in' argument: The type of the handle that the channel communicates           with, or 0 if there is no associated handle
 * @in_Handle: Used to pass an 'in' argument: The handle that the channel communicates with,           or 0 if there is no associated handle
 * @error: If not %NULL, used to return errors if %FALSE 
 *  is returned
 * @loop: If not %NULL, set before re-entering 
 *  the main loop, to point to a #GMainLoop 
 *  which can be used to cancel this call with 
 *  g_main_loop_quit(), causing a return of 
 *  %FALSE with @error set to %TP_DBUS_ERROR_CANCELLED
 *
 * Call the method HandleTube and run the main loop
 * until it returns. Before calling this method, you must
 * add a reference to any borrowed objects you need to keep,
 * and generally ensure that everything is in a consistent
 * state.
 *
 * Called when a tube handler should handle a new tube within a tubes channel.
 *
 * Returns: TRUE on success, FALSE and sets @error on error
 */
gboolean
emp_cli_tube_handler_run_handle_tube (TpProxy *proxy,
    gint timeout_ms,
    const gchar *in_Bus_Name,
    const gchar *in_Connection,
    const gchar *in_Channel,
    guint in_Handle_Type,
    guint in_Handle,
    GError **error,
    GMainLoop **loop)
{
  DBusGProxy *iface;
  GQuark interface = EMP_IFACE_QUARK_TUBE_HANDLER;
  TpProxyPendingCall *pc;
  _emp_cli_tube_handler_run_state_handle_tube state = {
      NULL /* loop */, error,
      FALSE /* completed */, FALSE /* success */ };

  g_return_val_if_fail (TP_IS_PROXY (proxy), FALSE);

  iface = tp_proxy_borrow_interface_by_id
       ((TpProxy *) proxy, interface, error);

  if (iface == NULL)
    return FALSE;

  state.loop = g_main_loop_new (NULL, FALSE);

  pc = tp_proxy_pending_call_v0_new ((TpProxy *) proxy,
      interface, "HandleTube", iface,
      _emp_cli_tube_handler_finish_running_handle_tube,
      NULL, &state, NULL, NULL, TRUE);

  if (loop != NULL)
    *loop = state.loop;

  tp_proxy_pending_call_v0_take_pending_call (pc,
      dbus_g_proxy_begin_call_with_timeout (iface,
          "HandleTube",
          _emp_cli_tube_handler_collect_callback_handle_tube,
          pc,
          tp_proxy_pending_call_v0_completed,
          timeout_ms,
              G_TYPE_STRING, in_Bus_Name,
              DBUS_TYPE_G_OBJECT_PATH, in_Connection,
              DBUS_TYPE_G_OBJECT_PATH, in_Channel,
              G_TYPE_UINT, in_Handle_Type,
              G_TYPE_UINT, in_Handle,
          G_TYPE_INVALID));

  if (!state.completed)
    g_main_loop_run (state.loop);

  if (!state.completed)
    tp_proxy_pending_call_cancel (pc);

  if (loop != NULL)
    *loop = NULL;

  g_main_loop_unref (state.loop);

  return state.success;
}


/*
 * emp_cli_misc_add_signals:
 * @self: the #TpProxy
 * @quark: a quark whose string value is the interface
 *   name whose signals should be added
 * @proxy: the D-Bus proxy to which to add the signals
 * @unused: not used for anything
 *
 * Tell dbus-glib that @proxy has the signatures of all
 * signals on the given interface, if it's one we
 * support.
 *
 * This function should be used as a signal handler for
 * #TpProxy::interface-added.
 */
static void
emp_cli_misc_add_signals (TpProxy *self G_GNUC_UNUSED,
    guint quark,
    DBusGProxy *proxy,
    gpointer unused G_GNUC_UNUSED)
{
  if (quark == EMP_IFACE_QUARK_DEBUG)
    emp_cli_add_signals_for_debug (proxy);
}
